#include "propellerloader.h"
#include "termios.h"
#include <QApplication>

PropellerLoader::PropellerLoader() : device(NULL), progress(NULL)
{

}

PropellerLoader::PropellerLoader(AbstractSerial *device, QProgressBar* progress)
{
    this->device = device;
    this->progress = progress;
}

PropellerLoader::LoaderResult PropellerLoader::updateFirmware(QByteArray data, bool useReset)
{

    // naive sanity checks
    if (data.size() < 10)  {
        return PropellerLoader::LoaderInvalidFile;
    }
    if (data.size() > 0x8000)  {
        return PropellerLoader::LoaderInvalidFile;
    }
    uint16_t binarySize = (uint8_t)data[8] + ((uint8_t)data[9] << 8);
    if (binarySize != data.size()) {
        return PropellerLoader::LoaderInvalidFile;
    }

    if (binarySize & 3) {
        return PropellerLoader::LoaderInvalidFile;
    }

    int res, i;

    // pack binary
    PropellerPacker packer;
    for (i = 0; i < data.size(); i++) {
        uint8_t byte = (uint8_t)data[i];
        for (int j = 0; j < 8; j++) {
            packer.pack(byte & 1);
            byte >>= 1;
        }
    }
    QByteArray packed = packer.finalize();

    progressValue = 0;
    progressMax = packed.size() + 500 + 30;

    if (useReset) { // hardware reset - service mode only
        device->setDtr(true);
        // do not need a pause here since we are using a USB adapter -
        // control requests are separated by millisecond interval
        device->setDtr(false);
    } else { // software reset - normal mode
        // send a single pulse to start bootloader
        device->putChar(-1);
    }

    {
        // allow some time for chip startup
        QTime t;
        t.start();
        while (t.elapsed() < 100) {};
    }

    sendCalibration();
    char bytes[258];

    LFSR = 'P';
    for (i = 0; i < 250; i++) {
        bytes[i] = (iterateLFSR() ? -1 : -2);
    }
    device->write(bytes, 250);
    advance(250);

    for (i = 0; i < 258; i++) {
        bytes[i] = 0xF9;
    }
    device->write(bytes, 250);

    int bytesRead = 0;
    char c[250];
    QTime t;
    t.start();
    while ((bytesRead < 250) && (t.elapsed() < 1000)) {
        res = device->read(c + bytesRead, 250 - bytesRead);
        if (res < 0) {
            return PropellerLoader::LoaderCommError;
        }
        advance(res);
        bytesRead += res;
    }

    for (i = 0; i < 250; i++) {
        uint8_t my = iterateLFSR();
        uint8_t reply = (c[i] == -1) ? 1 : 0;
        if (my != reply) {
            return PropellerLoader::LoaderCommError;
        }
    }

    bytesRead = 0;
    device->write(bytes + 250, 8);
    while ((bytesRead < 8) && (t.elapsed() < 1000)) {
        res = device->read(c + bytesRead, 8 - bytesRead);
        if (res < 0) {
            return PropellerLoader::LoaderCommError;
        }
        bytesRead += res;
    }
    if (bytesRead < 8) {
        return PropellerLoader::LoaderCommError;
    }

    uint8_t version = 0;
    for (i = 0; i < 8; i++) {
        version = (version >> 1) | ((c[i] == -1) ? 0x80 : 0);
    }
    if (version != 1) {
        return PropellerLoader::LoaderCommError;
    }

    // send command
    // always send 1 - load RAM and launch
    sendLong(1);
    // send count
    sendLong(binarySize >> 2);

    sendBytes(packed.data(), packed.size());

    char reply;
    if (!getReply(&reply)) {
        return PropellerLoader::LoaderCommError;
    }
    if (reply != -2) {
        return PropellerLoader::LoaderCommError;
    }


    return PropellerLoader::LoaderOK;
}

bool PropellerLoader::getReply(char *c)
{
    QTime t;
    int attempts = 30;
    while (attempts) {
        device->putChar(0xF9);
        t.start();
        while (t.elapsed() < 10) {};
        if (device->getChar(c)) {
            return true;
        }
        attempts--;
    }
    return false;
}

void PropellerLoader::sendByte(uint8_t byte)
{
    char buf[8];
    for (int i = 0; i < 8; i++) {
        if (byte & 1) buf[i] = -1;
        else buf[i] = -2;
    }
    device->write(buf, 8);
}

void PropellerLoader::sendLong(uint32_t num)
{
    char buf[32];
    for (int i = 0; i < 32; i++) {
        if (num & 1) buf[i] = -1;
        else buf[i] = -2;
        num >>= 1;
    }
    device->write(buf, 32);
}

void PropellerLoader::sendBytes(char* data, uint16_t count)
{
    int sent = 0;
    while (sent < count) {
        int toSend = count - sent;
        if (toSend > 2048) {
            toSend = 2048;
        }
        device->write(data + sent, toSend);
        advance(toSend);
        sent += toSend;
    }
}

void PropellerLoader::sendCalibration()
{
    device->putChar(0xF9);
}

uint8_t PropellerLoader::iterateLFSR()
{
    uint8_t bit = LFSR & 1;
    LFSR = (LFSR << 1) | (((LFSR >> 7) ^ (LFSR >> 5) ^ (LFSR >> 4) ^ (LFSR >> 1)) & 1);
    return bit;
}

void PropellerLoader::advance(int value)
{
    progressValue += value;
    if (progress) {
        progress->setMaximum(progressMax);
        progress->setValue(progressValue);
    }
    QApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
}

PropellerPacker::PropellerPacker() : bits(0xFFFFU), bitsUsed(0)
{

}

void PropellerPacker::pack(uint8_t bit)
{
    int bitsRequired;
    uint16_t mask;
    if (bit) {
        bitsRequired = 2;
        mask = 1;
    } else {
        bitsRequired = 3;
        mask = 3;
    }
    if (bitsRequired + bitsUsed > 10) {
        uint8_t byte = (bits >> 1) & 0xFF;
        bytes.append(byte);
        bitsUsed = 0;
        bits = 0xFFFFU;
    }
    mask = ~(mask << bitsUsed);
    bits &= mask;
    bitsUsed += bitsRequired;
}

QByteArray& PropellerPacker::finalize()
{
    if (bitsUsed) {
        uint8_t byte = (bits >> 1) & 0xFF;
        bytes.append(byte);
    }
    bitsUsed = 0;
    bits = 0xFFFFU;
    return bytes;
}
